import re

from mongoframes.factory import blueprints
from mongoframes.factory import makers
from mongoframes.factory import presets

from tests.fixtures import *


def test_blueprint_defaults():
    """
    Defining a `Blueprint` class should provide some defaults.
    """

    class DragonBlueprint(blueprints.Blueprint):

        _frame_cls = Dragon

    assert DragonBlueprint._frame_cls == Dragon
    assert DragonBlueprint._instructions == {}
    assert DragonBlueprint._meta_fields == set([])

def test_blueprint_assemble():
    """
    The `Blueprint.assemble` method should return an assembled a
    document/dictionary for the blueprint.
    """

    # Configure the blueprint
    class DragonBlueprint(blueprints.Blueprint):

        _frame_cls = Dragon
        _meta_fields = {'dummy_prop'}

        name = makers.Static('Burt')
        dummy_prop = makers.Static('foo')

    # Build a list if presets
    my_presets = [presets.Preset('breed', makers.Static('Fire-drake'))]

    # Check the assembled output of the blueprint is as expected
    assembled = DragonBlueprint.assemble(my_presets)

    assert assembled == {
        'breed': 'Fire-drake',
        'dummy_prop': 'foo',
        'name': 'Burt'
        }

def test_blueprint_finish():
    """
    The `Blueprint.finish` method should return a finished document/dictionary
    and meta document/dictionary for the blueprint.
    """

    # Configure the blueprint
    class DragonBlueprint(blueprints.Blueprint):

        _frame_cls = Dragon
        _meta_fields = {'dummy_prop'}

        name = makers.Static('Burt')
        dummy_prop = makers.Static('foo')

    # Build a list if presets
    my_presets = [presets.Preset('breed', makers.Static('Fire-drake'))]

    # Check the finished output of the blueprint is as expected
    finished, meta_finished = DragonBlueprint.finish(
        DragonBlueprint.assemble(my_presets),
        my_presets
        )

    assert finished == {'breed': 'Fire-drake', 'name': 'Burt'}
    assert meta_finished == {'dummy_prop': 'foo'}

def test_blueprint_reset(mocker):
    """
    The `Blueprint.reset` method should call the reset of all makers in the
    blueprints instructions and any in the given list of presets.
    """

    # Configure the blueprint
    class DragonBlueprint(blueprints.Blueprint):

        _frame_cls = Dragon
        _meta_fields = {'dummy_prop'}

        name = makers.Static('Burt')
        dummy_prop = makers.Static('foo')

    # Build a list if presets
    my_presets = [presets.Preset('breed', makers.Static('Fire-drake'))]

    # Spy on the maker reset methods
    mocker.spy(my_presets[0].maker, 'reset')
    mocker.spy(DragonBlueprint._instructions['name'], 'reset')
    mocker.spy(DragonBlueprint._instructions['dummy_prop'], 'reset')

    # Reset the blueprint
    DragonBlueprint.reset(my_presets)

    # Check each maker reset method was called
    assert my_presets[0].maker.reset.call_count == 1
    assert DragonBlueprint._instructions['name'].reset.call_count == 1
    assert DragonBlueprint._instructions['dummy_prop'].reset.call_count == 1


__all__ = ['Preset']


class Preset:
    """
    Presets match field names to makers (any function that can produce a value
    for a field). They provide default makers for fields that are not directly
    defined in a `Blueprints` instructions.

    The pattern used to match a preset with a maker should be a compiled regular
    expression or a string. If a string is provided then the match must match
    exactly.
    """

    def __init__(self, pattern, maker):

        # The pattern used to associate a field name with a maker
        self._pattern = pattern

        # The maker that will be used to produce values for this field
        self._maker = maker

    # Read-only properties

    @property
    def maker(self):
        return self._maker

    @property
    def pattern(self):
        return self._pattern

    # Public methods

    def match(self, field_name):
        """
        Return True if the given field name matches the preset's pattern.
        """

        if hasattr(self._pattern, 'match'):
            return self._pattern.match(field_name) is not None

        return self._pattern == field_name

    # Class methods

    @classmethod
    def find(cls, presets, field_name):
        """
        Search a list of presets and return the first to match the specified
        field name or None if there are not matches.
        """
        for preset in presets:
            if preset.match(field_name):
                return preset


import copy

__all__ = ['Factory']


class Factory:
    """
    The `Factory` class is responsible for production of fake data for.
    Production of fake data is a two stage process:

    Assembly (see `assemble`)
    :   A `Quota` of documents is assembled based on a `Blueprint`.

        At this stage the documents contain a mixture of static and dynamic
        data. Dynamic data is data that will be transformed during population,
        for example a field might contain a value of `'now,tomorrow'` which on
        population will be converted to a date/time between now and tomorrow.

        Once assembled the generated documents are returned as a list and can be
        either used immediately to populate the database or saved out as a
        template for populating the database in future (for example when
            building a set of test data).

    Population (see `populate`)
    :   A database is populated based on a `Blueprint` and pre-assembled list of
        documents.

        During this stage dynamic data is converted to static data suitable for
        inserting in the database (this process is call finishing).

        Prior to and after inserting the a document into the database the
        `factory_insert` and `factory_inserted` events are triggered to allow
        `Frame` classes to modify the insert behaviour for factories.
    """

    def __init__(self, presets=None):
        # A list of presets for the factory
        self._presets = presets or []

    # Read-only properties

    @property
    def presets(self):
        return self._presets

    # Public methods

    def assemble(self, blueprint, quota):
        """Assemble a quota of fake documents"""

        # Reset the blueprint
        blueprint.reset()

        # Assemble the documents
        documents = []
        for i in range(0, int(quota)):
            documents.append(blueprint.assemble(self.presets))

        return documents

    def finish(self, blueprint, documents):
        """Apply finishing to a list of pre-assembled documents"""

        # Reset the blueprint
        blueprint.reset()

        # Finish the documents
        finished = []
        for document in documents:
            finished.append(blueprint.finish(document, self.presets))

        return finished

    def populate(self, blueprint, documents):
        """Populate the database with fake documents"""

        # Finish the documents
        documents = self.finish(blueprint, documents)

        # Convert the documents to frame instances
        frames = []
        for document in documents:
            [frame_document, meta_document] = document

            # Initialize the frame
            frame = blueprint.get_frame_cls()(frame_document)

            # Apply any meta fields
            for key, value in meta_document.items():
                setattr(frame, key, value)

            frames.append(frame)

        # Insert the documents
        blueprint.on_fake(frames)
        frames = blueprint.get_frame_cls().insert_many(frames)
        blueprint.on_faked(frames)

        return frames


import re

from mongoframes.factory import makers
from mongoframes.factory import presets


def test_preset_read_only_props():
    """
    The `Preset` class read-only properties should return the correct values.
    """

    maker = makers.Static('bar')
    preset = presets.Preset('foo', maker)

    # Check the read-only properties of the preset return the correct values
    assert preset.pattern == 'foo'
    assert preset.maker == maker

def test_preset_match():
    """
    The `Preset.match` method should return true if a field name matches a
    pattern.
    """

    # Configure the preset with a simple string match
    preset = presets.Preset('foo', makers.Static('bar'))
    assert preset.match('foo') == True
    assert preset.match('bar') == False

    # Configure the preset with a regular expression match
    preset = presets.Preset(re.compile('foo.*'), makers.Static('bar'))
    assert preset.match('foobar') == True
    assert preset.match('barfoo') == False

def test_preset_find():
    """
    The `Preset.find` class method should return the first preset that matches
    the given field name.
    """
    my_presets = [
        presets.Preset('foo', makers.Static('foo')),
        presets.Preset('bar', makers.Static('bar')),
        presets.Preset('zee', makers.Static('zee')),
        presets.Preset(re.compile('foo.*'), makers.Static('foo*')),
        presets.Preset(re.compile('bar.*'), makers.Static('bar*')),
        presets.Preset(re.compile('zee.*'), makers.Static('zee*'))
        ]

    # Check that `Preset.find` returns the expected presets
    assert presets.Preset.find(my_presets, 'foo') == my_presets[0]
    assert presets.Preset.find(my_presets, 'bar') == my_presets[1]
    assert presets.Preset.find(my_presets, 'zee') == my_presets[2]
    assert presets.Preset.find(my_presets, 'foo1') == my_presets[3]
    assert presets.Preset.find(my_presets, 'bar2') == my_presets[4]
    assert presets.Preset.find(my_presets, 'zee3') == my_presets[5]